### # #### # ##### ######### ## ###### ##### ### ##### ########
##
## Copyright (c) 2005-2010, SkeTo Project
## All rights reserved.
##
## This file is part of the SkeTo Library. This library is free
## software; you can redistribute it and/or modify it under the
## terms of the MIT License (see COPYING in the package).
## 
## This library is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
##
### # #### # ##### ######### ## ###### ##### ### ##### ########

AC_DEFUN([CHECK_MPICXX], [
AC_PREREQ(2.52)

# use MPI at the given path
# mpi_base is set to the given path
# mpi_bin is set to mpi_base/bin
# mpi_lib is set to mpi_base/lib
AC_ARG_WITH(mpi,[using specific MPI (base path)],[mpi_base="${with_mpi%%/}/"; mpi_bin="${mpi_base}bin/"; mpi_lib="${mpi_base}lib/"; ])

AC_REQUIRE([AC_PROG_CXX])
# when MPICXX is set (exported) in the environment, use it
AC_ARG_VAR(MPICXX,[MPI C++ compiler])
if test x = x$mpi_bin; then
  # search the compiler at the usual path
  # the compilers listed in the second argumnt will be tested in the order.
  AC_CHECK_PROGS(MPICXX, mpicxx mpic++ mpiCC, $CXX)
else
  # search the compiler at the given path
  AC_PATH_PROGS(MPICXX, mpicxx mpic++ mpiCC, $CXX, ${mpi_bin})
fi
AC_SUBST(MPICXX)

# the same as MPICXX
AC_ARG_VAR(MPIRUN,[MPI program launcher])
if test x = x$mpi_bin; then
  AC_CHECK_PROGS(MPIRUN, mpirun mpiexec,"")
else
  AC_PATH_PROGS(MPIRUN, mpirun mpiexec,"", ${mpi_bin})
fi
AC_SUBST(MPIRUN)

mpi_save_CXX="$CXX"
CXX="$MPICXX"

mpi_h_ok="yes"
mpi_libs_ok="yes"

# check whether mpi.h is available or not
AC_MSG_CHECKING([for mpi.h])
AC_TRY_COMPILE([#include <mpi.h>],
			   [],
               [AC_MSG_RESULT(yes)], 
               [mpi_h_ok="no"; AC_MSG_RESULT(no)])

mpi_save_LIBS="$LIBS"
if test x != x$mpi_lib; then
   LIBS="-L$mpi_lib "
else
   LIBS=""
fi

# checks whether MPI_Init is correctly linked or not.
# it first tests without options, 
# then it tests with -lXXX for XXX listed in the second argumnet
AC_SEARCH_LIBS(MPI_Init, 
               mpich mpi, 
               [], 
               [mpi_libs_ok="no"])
MPILIBS="$LIBS"
AC_SUBST(MPILIBS)

LIBS="$mpi_save_LIBS"

CXX="$mpi_save_CXX"

# if all is available, set the variable HAVE_MPI in config.h
# otherwise stop the configure script.
if test xyesyes = x"$mpi_libs_ok$mpi_h_ok"; then
   AC_DEFINE(HAVE_MPI,1,[Whether you have MPI.])
else
   echo "!! MPI is required!                                              !!"
   echo "!! If you have installed MPI, please specify that using options. !!"
   exit
fi
])

