/*** * **** * ***** ********* ** ****** ***** *** ***** ********
**
** Copyright (c) 2005-2010, SkeTo Project
** All rights reserved.
**
** This file is part of the SkeTo Library. This library is free
** software; you can redistribute it and/or modify it under the
** terms of the MIT License (see COPYING in the package).
** 
** This library is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
**
*** * **** * ***** ********* ** ****** ***** *** ***** ********/

/**
 *
 * @file list_skeletons_predef.h
 * @brief A set of predefinition of class names, etc.
 */
#ifndef __LIST_SKELETONS_PREDEF_H__
#define __LIST_SKELETONS_PREDEF_H__

namespace sketo {
  template <typename A> class dist_list;

  namespace _impl {
    template <typename A> class dist_list_buffer;

    enum {
      TAG_REDUCE,
      TAG_SCAN1,
      TAG_PRINT,
      TAG_SHIFTL,
      TAG_SHIFTR,
      TAG_GATHER,
    };
    
    class ls_adpt {
    public:
      template <typename A, typename OP>
      static void
      global_reduce(const OP &oplus,
		    A* pval);

      template <typename A, typename OP>
      static void
      global_scan(const OP &oplus,
		  A* pfirst,
		  A* ptotal);

      template <typename A, typename OP>
      static void
      global_scanr(const OP &oplus,
		   A* plast,
		   A* ptotal);

#ifdef __SKETO_NO_FUSION__
      template <typename F, typename A>
      static dist_list<A>
      generate(int global_size, const F& f);

      template <typename F, typename A, typename B>
      static dist_list<B>
      map(const F& f,
	  const dist_list<A>& as);

      template <typename F, typename A, typename B>
      static dist_list<B>
      map_with_index(const F& f,
		     const dist_list<A>& as);
      
      template <typename F, typename A, typename B, typename C>
      static dist_list<C>
      zipwith(const F& f,
	      const dist_list<A>& as,
	      const dist_list<B>& bs);
      
      template <typename OPLUS, typename A>
      static A
      reduce(const OPLUS &oplus,
	     const dist_list<A> &as);

      template <typename OPLUS, typename A>
      static dist_list<A>
      scan(const OPLUS& oplus,
	   A e,
	   const dist_list<A> &as,
	   A* last_val);

      template <typename OPLUS, typename A>
      static dist_list<A>
      scanr(const OPLUS& oplus,
	    A e,
	    const dist_list<A> &as,
	    A* first_val);

      template <typename A>
      static dist_list<A>
      shiftl(const A& rightmost,
	     const dist_list<A> &as,
	     A* leftmost);

      template <typename A>
      static dist_list<A>
      shiftr(const A& leftmost,
	     const dist_list<A> &as,
	     A* rightmost);

#else
      template <typename OPLUS, typename AS, typename A>
      static A
      reduce(const OPLUS &op, const AS &as);
      
      template <typename OP, typename AS, typename A>
      static dist_list<A>*
      scanobj_ow_impl(dist_list<A>* bs, const OP op, const AS as, const A e, A* last);

      template <typename OP, typename AS, typename A>
      static void
      scanobj_subeval(dist_list<A>* bs, A* first_val, const OP op, const AS as, const A e, A* last);
      
      template <typename OP, typename AS, typename A>
      static dist_list<A>*
      scanrobj_ow_impl(dist_list<A>* bs, const OP op, const AS as, const A e, A* last);
      
      template <typename OP, typename AS, typename A>
      static void
      scanrobj_subeval(dist_list<A>* bs, A* last_val, const OP op, const AS as, const A e, A* first);

      template <typename A>
      static void
      shiftl_edge(A old_leftmost,
		  A* new_rightmost);

      template <typename AS, typename A>
      static dist_list<A>*
      shiftlobj_ow_impl(dist_list<A>* bs, const AS as, const A fromR);

      template <typename A>
      static void
      shiftr_edge(A old_rightmost,
		  A* new_leftmost);

      template <typename AS, typename A>
      static dist_list<A>*
      shiftrobj_ow_impl(dist_list<A>* bs, const AS as, const A fromL);
      
#endif

      // For the following four skeletons, we have not
      // implemented fused implementation yet.
      template <typename OPLUS, typename A>
      static dist_list<A>
      postscan(const OPLUS &oplus,
	       const dist_list<A> &as);

      template <typename OPLUS, typename A>
      static dist_list<A>
      postscanr(const OPLUS &oplus,
		const dist_list<A> &as);

      template <typename F, typename G, typename OMINUS, typename OPLUS,
		typename A, typename B, typename C>
      static dist_list<B>
      gscanl(const F &f,
	     const B &c,
	     const G &g,
	     const OMINUS &ominus,
	     const OPLUS &oplus,
	     const dist_list<A> &as);
	
      template <typename F, typename G, typename OMINUS, typename OPLUS,
		typename A, typename B, typename C>
      static dist_list<B>
      gscanr(const F &f,
	     const B &c,
	     const G &g,
	     const OMINUS &ominus,
	     const OPLUS &oplus,
	     const dist_list<A>& as);

    };

}

}




#endif // __LIST_SKELETONS_PREDEF_H__
