/*** * **** * ***** ********* ** ****** ***** *** ***** ********
 **
 ** Copyright (c) 2005-2010, SkeTo Project
 ** All rights reserved.
 **
 ** This file is part of the SkeTo Library. This library is free
 ** software; you can redistribute it and/or modify it under the
 ** terms of the MIT License (see COPYING in the package).
 ** 
 ** This library is distributed in the hope that it will be useful,
 ** but WITHOUT ANY WARRANTY; without even the implied warranty of
 ** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
 **
 *** * **** * ***** ********* ** ****** ***** *** ***** ********/

#ifndef __LIST_SKELETONS_WITH_FUSION_TPP__
#define __LIST_SKELETONS_WITH_FUSION_TPP__

#include "list_skeletons_common.tpp"

#endif // __LIST_SKELETONS_WITH_FUSION_TPP__

namespace sketo {
  namespace _impl {

    //------------------------------------------------------------
    // Rules for Owerwriting
    
    template <typename F>
    template <typename T>
    void
    ls_genobj<F>::check_overwrite(_impl::dist_list_buffer<T>*,
                                  int *,
                                  bool *,
                                  bool *,
                                  bool *scan_shift) const
    {
      *scan_shift = false;
    }

    template <typename F, typename AS>
    template <typename T>
    void
    ls_mapobj<F, AS>::check_overwrite(_impl::dist_list_buffer<T>* target,
                                      int *refcount,
                                      bool *no_deps,
                                      bool *child_no_deps,
                                      bool *scan_shift) const
    {
      as.check_overwrite(target, refcount, no_deps, child_no_deps, scan_shift);
      *scan_shift = false;
    }

    template <typename F, typename AS, typename BS>
    template <typename T>
    void
    ls_zipwobj<F, AS, BS>::check_overwrite(_impl::dist_list_buffer<T>* target,
                                           int *refcount,
                                           bool *no_deps,
                                           bool *child_no_deps,
                                           bool *scan_shift) const
    {
      as.check_overwrite(target, refcount, no_deps, child_no_deps, scan_shift);
      bs.check_overwrite(target, refcount, no_deps, child_no_deps, scan_shift);
      *scan_shift = false;
    }

    template <typename AS, typename A>
    template <typename T>
    void
    ls_shiftlobj<AS, A>::check_overwrite(_impl::dist_list_buffer<T>* target,
                                         int *refcount,
                                         bool *no_deps,
                                         bool *child_no_deps,
                                         bool *scan_shift) const
    {
      int refcount2 = 0;
      bool no_deps2 = true;
      as.check_overwrite(target, &refcount2, &no_deps2, child_no_deps, scan_shift);
      *refcount += refcount2;
      *no_deps = *no_deps && no_deps2;
      *child_no_deps = no_deps2;
      *scan_shift = true;
    }

    template <typename AS, typename A>
    template <typename T>
    void
    ls_shiftrobj<AS, A>::check_overwrite(_impl::dist_list_buffer<T>* target,
                                         int *refcount,
                                         bool *no_deps,
                                         bool *child_no_deps,
                                         bool *scan_shift) const
    {
      int refcount2 = 0;
      bool no_deps2 = true;
      as.check_overwrite(target, &refcount2, &no_deps2, child_no_deps, scan_shift);
      *refcount += refcount2;
      *no_deps = *no_deps && no_deps2;
      *child_no_deps = no_deps2;
      *scan_shift = true;
    }

    template <typename OP, typename AS, typename A>
    template <typename T>
    void ls_scanobj<OP, AS, A>::check_overwrite(_impl::dist_list_buffer<T>* target,
                                                int *refcount,
                                                bool *no_deps,
                                                bool *child_no_deps,
                                                bool *scan_shift) const
    {
      int refcount2 = 0;
      bool no_deps2 = true;
      as.check_overwrite(target, &refcount2, &no_deps2, child_no_deps, scan_shift);
      *refcount += refcount2;
      *no_deps = *no_deps && no_deps2;
      *child_no_deps = no_deps2;
      *scan_shift = true;
    }

    template <typename OP, typename AS, typename A>
    template <typename T>
    void ls_scanrobj<OP, AS, A>::check_overwrite(_impl::dist_list_buffer<T>* target,
                                                 int *refcount,
                                                 bool *no_deps,
                                                 bool *child_no_deps,
                                                 bool *scan_shift) const
    {
      int refcount2 = 0;
      bool no_deps2 = true;
      as.check_overwrite(target, &refcount2, &no_deps2, child_no_deps, scan_shift);
      *refcount += refcount2;
      *no_deps = *no_deps && no_deps2;
      *child_no_deps = no_deps2;
      *scan_shift = true;
    }
    
    //------------------------------------------------------------

    
    template <typename OPLUS, typename AS, typename A>
    A
    ls_adpt::reduce(const OPLUS &oplus, const AS &as)
    {
      // local reduction
      A result = as.local_get(0);
      {
        const int n = as.get_local_size();
        for (int i = 1; i < n; ++i) {
          result = oplus(result, as.local_get(i));
        }
      }
      
      _impl::ls_adpt::global_reduce(oplus, &result);
      return result;
    }

    template <typename OP, typename AS, typename A>
    dist_list<A>*
    ls_adpt::scanobj_ow_impl(dist_list<A>* bs, const OP op, const AS as, const A e, A* last)
    {
      // local reduce
      A total_sum;
      {
        const int n = as.get_local_size();
        total_sum = as.local_get(0);
        for (int i = 1; i < n; ++i) {
          total_sum = op(total_sum, as.local_get(i));
        }
      }
      
      // global scan
      A first_val;
      global_scan(op, &first_val, &total_sum);
      if (rank == 0) {
        first_val = e;
      } else {
        first_val = op(e, first_val);
      }
      
      // share the last value (= total_sum)
      if (last) {
        if (sketo::rank == 0) *last = op(e, total_sum);
        _impl::transporter::bcast<A>(last, 1, 0);
      }
      
      // local scan
      {
        A* const bs_buf = bs->data->buf;
        const int n = as.get_local_size();
        A orgdata = as.local_get(0);
        bs_buf[0] = first_val;
        for (int i = 1; i < n; ++i) {
          A temp = op(bs_buf[i-1], orgdata);
          orgdata = as.local_get(i);
          bs_buf[i] = temp;
        }
      }
      return bs;
    }

    template <typename OP, typename AS, typename A>
    void
    ls_adpt::scanobj_subeval(dist_list<A>* bs, A* first_val, const OP op, const AS as, const A e, A* last)
    {
      // local scan
      A total_sum;
      {
        const int n = as.get_local_size();
        total_sum = identity_element(op);
        A* const bs_buf = bs->data->buf;
        for (int i = 0; i < n; ++i) {
          bs_buf[i] = total_sum;
          total_sum = op(total_sum, as.local_get(i));
        }
      }

      // global_scan
      global_scan(op, first_val, &total_sum);
      if (rank == 0) {
        *first_val = e;
      } else {
        *first_val = op(e, *first_val);
      }

      // share the last value
      if (last) {
        if (sketo::rank == 0) *last = op(e, total_sum);
        _impl::transporter::bcast<A>(last, 1, 0);
      }
    }

    template <typename OP, typename AS, typename A>
    void
    ls_adpt::scanrobj_subeval(dist_list<A>* bs, A* last_val, const OP op, const AS as, const A e, A* first)
    {
      // local scanr
      A total_sum;
      {
        const int n = as.get_local_size();
        total_sum = identity_element(op);
        A* const bs_buf = bs->data->buf;
        for (int i = n - 1; i >= 0; --i) {
          bs_buf[i] = total_sum;
          total_sum = op(as.local_get(i), total_sum);
        }
      }

      // global_scanr
      global_scanr(op, last_val, &total_sum);
      if (rank == procs - 1) {
        *last_val = e;
      } else {
        *last_val = op(*last_val, e);
      }

      // share the first value
      if (first) {
        if (sketo::rank == procs - 1) *first = op(total_sum, e);
        _impl::transporter::bcast<A>(first, 1, procs-1);
      }
    }
      
    
    template <typename OP, typename AS, typename A>
    dist_list<A>*
    ls_scanobj<OP, AS, A>::ow_impl(dist_list<A>* bs) const
    {
      return _impl::ls_adpt::scanobj_ow_impl(bs, op, as, e, last);
    }

    template <typename OP, typename AS, typename A>
    dist_list<A>*
    ls_adpt::scanrobj_ow_impl(dist_list<A>* bs, const OP op, const AS as, const A e, A* first)
    {
      // local reduce
      A total_sum;
      {
        const int n = as.get_local_size();
        total_sum = as.local_get(0);
        for (int i = 1; i < n; ++i) {
          total_sum = op(total_sum, as.local_get(i));
        }
      }
    
      // global scanr
      A last_val;
      global_scanr(op, &last_val, &total_sum);
      if (rank == procs - 1) {
        last_val = e;
      } else {
        last_val = op(last_val, e);
      }
    
      // share the first value (= total_sum)
      if (first) {
        if (sketo::rank == procs - 1) *first = op(total_sum, e);
        _impl::transporter::bcast<A>(first, 1, procs-1);
      }

      // local scan
      {
        A* const bs_buf = bs->data->buf;
        bs_buf[as.get_local_size() - 1] = last_val;
        for (int i = as.get_local_size() - 2; i >= 0; --i) {
          bs_buf[i] = op(as.local_get(i+1), bs_buf[i+1]);
        }
      }
    
      return bs;
    }

    template <typename OP, typename AS, typename A>
    dist_list<A>*
    ls_scanrobj<OP, AS, A>::ow_impl(dist_list<A>* bs) const
    {
      return _impl::ls_adpt::scanrobj_ow_impl(bs, op, as, e, first);
    }
    
    template <typename A>
    void
    _impl::ls_adpt::shiftl_edge(A old_leftmost,
                                A* new_rightmost)
    {
      simple_send_request<A> reqs;
      simple_recv_request<A> reqr;
      // sending or receiving edges
      if (sketo::rank != 0) {
        _impl::transporter::isend(&old_leftmost, 1, sketo::rank - 1,
                                  TAG_SHIFTL, &reqs);
      }
      if (sketo::rank != sketo::procs - 1) {
        _impl::transporter::irecv(new_rightmost, 1, sketo::rank + 1,
                                  TAG_SHIFTL, &reqr);
      }
      
      // wait sening or receiving
      if (sketo::rank != 0) {
        reqs.wait();
      }
      if (sketo::rank != sketo::procs - 1) {
        reqr.wait();
      }

    }

    template <typename AS, typename A>
    dist_list<A>*
    ls_adpt::shiftlobj_ow_impl(dist_list<A>* bs, const AS as, const A fromR)
    {
      for (int i = 0; i < as.get_local_size() - 1; ++i) {
        bs->data->buf[i] = as.local_get(i + 1);
      }
      bs->data->buf[as.get_local_size() - 1] = fromR;
      return bs;
    }
	
    template <typename AS, typename A>
    dist_list<A>*
    ls_shiftlobj<AS, A>::ow_impl(dist_list<A>* bs) const
    {
      return _impl::ls_adpt::shiftlobj_ow_impl(bs, as, fromR);
    }

    template <typename A>
    void
    _impl::ls_adpt::shiftr_edge(A old_rightmost,
                                A* new_leftmost)
    {
      simple_send_request<A> reqs;
      simple_recv_request<A> reqr;
      // sending or receiving edges
      if (sketo::rank != sketo::procs - 1) {
        _impl::transporter::isend(&old_rightmost, 1, sketo::rank + 1,
                                  TAG_SHIFTR, &reqs);
      }
      if (sketo::rank != 0) {
        _impl::transporter::irecv(new_leftmost, 1, sketo::rank - 1,
                                  TAG_SHIFTR, &reqr);
      }
      
      // wait sening or receiving
      if (sketo::rank != sketo::procs - 1) {
        reqs.wait();
      }
      if (sketo::rank != 0) {
        reqr.wait();
      }

    }

    template <typename AS, typename A>
    dist_list<A>*
    ls_adpt::shiftrobj_ow_impl(dist_list<A>* bs, const AS as, const A fromL)
    {
      for (int i = as.get_local_size() - 1; i >= 1; --i) {
        bs->data->buf[i] = as.local_get(i - 1);
      }
      bs->data->buf[0] = fromL;
      return bs;
    }
	
    template <typename AS, typename A>
    dist_list<A>*
    ls_shiftrobj<AS, A>::ow_impl(dist_list<A>* bs) const
    {
      return _impl::ls_adpt::shiftrobj_ow_impl(bs, as, fromL);
    }
    
    
  }
}
