/*** * **** * ***** ********* ** ****** ***** *** ***** ********
**
** Copyright (c) 2005-2010, SkeTo Project
** All rights reserved.
**
** This file is part of the SkeTo Library. This library is free
** software; you can redistribute it and/or modify it under the
** terms of the MIT License (see COPYING in the package).
** 
** This library is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
**
*** * **** * ***** ********* ** ****** ***** *** ***** ********/

/**
 * @file sketo_io.h
 * @brief Some utility functions for IO.
 * @author Kiminori MATSUAZKI, Kento EMOTO
 */

#ifndef __SKETO_IO_H__
#define __SKETO_IO_H__

namespace sketo {

  /**
   * @brief This output stream only write data from a single process
   * (the root process).
   */
  class single_ostream : public std::ofstream {
  public:
    // for iomanip
#if __GNUC__ > 2
    single_ostream&
    operator<<(__ostream_type& (*__pf)(__ostream_type&)){
      if (rank==0) std::cout << __pf;
      return *this;
    }
    single_ostream&
    operator<<(__ios_type& (*__pf)(__ios_type&)){
      if (rank==0) std::cout << __pf;
      return *this;
    }
    single_ostream&
    operator<<(ios_base& (*__pf) (ios_base&)){
      if (rank==0) std::cout << __pf;
      return *this;
    }
#else
    single_ostream&
    operator<<(ios& (*__pf) (ios&)){
      if (rank==0) std::cout << __pf;
      return *this;
    }
#endif

    // for other types
    template <typename C>
    single_ostream& operator<<(const C&v){
      if (rank==0) std::cout << v;
      return *this;
    }
  };
  extern single_ostream cout;

  void mprintf(const char* format, ...);
}

#endif // __SKETO_IO_H__
