/*** * **** * ***** ********* ** ****** ***** *** ***** ********
**
** Copyright (c) 2005-2010, SkeTo Project
** All rights reserved.
**
** This file is part of the SkeTo Library. This library is free
** software; you can redistribute it and/or modify it under the
** terms of the MIT License (see COPYING in the package).
** 
** This library is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
**
*** * **** * ***** ********* ** ****** ***** *** ***** ********/

/**
 * @file correlation.cpp
 * @brief This program compute correlation coefficient.
 * @author Yoshiki AKASHI (yoshiki@ipl.cs.uec.ac.jp)
 */

#include <cstdlib>
#include <cmath>
#include <ctime>
#include <sketo/list_skeletons.h>

struct Gen : public sketo::functions::base<double(int)> {
  double operator()(int) const {
    return static_cast< double >( std::rand( ) % 100 );
  }
} gen;

  
int sketo::main(int argc, char ** argv)
{
  if ( argc < 3 ) {
    sketo::cout << "usage: " << argv[ 0 ] << " size loop" << std::endl;
    return 0;
  }

  const int size = std::atoi( argv[ 1 ] );
  const int loop = std::atoi( argv[ 2 ] );

  std::srand( std::time( NULL ) ); // initialize random seed

//   skeleton::time::initialize( ); // for time ekeeping

  for ( int i = 0; i < loop; i++ ) {
    // generate data
    sketo::dist_list<double> xs = sketo::list_skeletons::generate(size, gen);
    sketo::dist_list<double> ys = sketo::list_skeletons::generate(size, gen);
    
    // calculate average
    double ave_x = sketo::list_skeletons::reduce(std::plus<double>(), xs) / size;
    double ave_y = sketo::list_skeletons::reduce(std::plus<double>(), ys) / size;
    
    // update xs with (x[i] - ave_x), ys with (y[i] - ave_y)
    xs = sketo::list_skeletons::map(std::bind2nd(std::minus<double>(), ave_x), xs);
    ys = sketo::list_skeletons::map(std::bind2nd(std::minus<double>(), ave_y), ys);
    
    // sigma (x - ave_x)^2
    double sum_x_x = sketo::list_skeletons::reduce(std::plus<double>(),
	               sketo::list_skeletons::map(sketo::functions::square<double>(), xs));
    // sigma (y - ave_y)^2
    double sum_y_y = sketo::list_skeletons::reduce(std::plus<double>(),
	               sketo::list_skeletons::map(sketo::functions::square<double>(), ys));
    // sigma (x - ave_x)(y - ave_y)
    double sum_x_y = sketo::list_skeletons::reduce(std::plus<double>(),
                       sketo::list_skeletons::zipwith(std::multiplies<double>(), xs, ys));
    
    // calculate correlation coefficient
    double r = sum_x_y / std::sqrt(sum_x_x * sum_y_y);

    // output result    
    sketo::cout << "correlation coefficient is " << r << std::endl;
  }

//   skeleton::time::set_time( "calculate" ); // for time ekeeping
//   skeleton::time::print( ); // show the result of time ekeeping
  
  return 0;
}
