/*** * **** * ***** ********* ** ****** ***** *** ***** ********
**
** Copyright (c) 2005-2010, SkeTo Project
** All rights reserved.
**
** This file is part of the SkeTo Library. This library is free
** software; you can redistribute it and/or modify it under the
** terms of the MIT License (see COPYING in the package).
** 
** This library is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
**
*** * **** * ***** ********* ** ****** ***** *** ***** ********/

/**
 * @file var-lambda.cpp
 * @brief An example program for lambda expressions
 * @author Kento EMOTO (emotoi@ipl.cs.uec.ac.jp)
 * @author Kiminori MATSUZAKI (kmatsu@ipl.t.u-tokyo.ac.jp)
 */

#include <cstdlib>
#include <ctime>
#include <sketo/list_skeletons.h>

using namespace sketo::list_skeletons;

int sketo::main(int argc, char** argv)
{
  if (argc < 2) {
    sketo::cout << "usage: " << argv[ 0 ] << " size " << std::endl;
    return 0;
  }

  const int size = std::atoi(argv[1]);

  // generate data
  dist_list<double> da =
    generate(size,
             [](int){return (double)(std::rand() % 100);});
  
  
  // calculate average
  auto plus = [](double a, double b){return a + b;};
  double ave = reduce(plus, da) / size;
  
  // calculate variance with fusions
  auto db = map([ave](double a){return a - ave;}, da);
  auto dc = map([](double a){return a * a;}, db);
  double var = reduce(plus, dc) / size;

  // output results
  sketo::cout << "average: "  << ave << std::endl;
  sketo::cout << "variance: " << var << std::endl;
  return 0;
}
