/*** * **** * ***** ********* ** ****** ***** *** ***** ********
**
** Copyright (c) 2005-2010, SkeTo Project
** All rights reserved.
**
** This file is part of the SkeTo Library. This library is free
** software; you can redistribute it and/or modify it under the
** terms of the MIT License (see COPYING in the package).
** 
** This library is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
**
*** * **** * ***** ********* ** ****** ***** *** ***** ********/
 
/**
 * @file sketo.cpp
 * @brief The skeleton common file.
 * @author Kiminori MATSUZAKI (kmatus@ipl.t.u-tokyo.ac.jp)
 * @author Yoshiki AKASHI (yoshiki@ipl.cs.uec.ac.jp)
 * @author Kento EMOTO (emoto@ipl.t.u-tokyo.ac.jp)
 */

#include "../include/sketo/sketo.h"

namespace sketo {
  int rank;  // MPI rank
  int procs; // Number of processors
  single_ostream cout;

  MPI_Comm context_comm;
  
  void init(int &argc, char **&argv) {
    MPI_Init(&argc, &argv);
    MPI_Comm_rank(MPI_COMM_WORLD, &rank);
    MPI_Comm_size(MPI_COMM_WORLD, &procs);
  }

  void finalize() {
    MPI_Finalize();
  }

  void mprintf(const char* format, ...) {
    if (!is_master()) return;
    va_list args;
    va_start(args, format);
    vprintf(format, args);
    va_end(args);
  }

};

int main(int argc, char **argv) {
  sketo::init(argc, argv);

  int retval = 0;
  try {
    retval = sketo::main(argc, argv);
  } catch (...) {
    std::cerr << "Something Wrong" << std::endl;
  }

  sketo::finalize();
  return retval;
}
